#include <PWMFreak.h>

#define X_STEP_PIN        5
#define Y_STEP_PIN        6
#define X_DIR_PIN         4
#define Y_DIR_PIN         7
#define ALL_POWER_PIN     8
#define ALL_POWER_ON      LOW
#define ALL_POWER_OFF     HIGH
#define LASER_POWER_PIN   10
#define LASER_POWER_OFF   LOW

#define PIN_FILTER_COUNT  10
//#define FIX_Z_DIST        100

//#define EMERGENCY_PIN     9
//#define EMERGENCY_ACTIVE  LOW

#define X_LEFT_POS_PIN    A0
#define X_LEFT_POS_ACTIVE LOW
#define X_RIGHT_POS_PIN   A1
#define X_RIGHT_POS_ACTIVE LOW
#define Y_TOP_POS_PIN     A2
#define Y_TOP_POS_ACTIVE  LOW
#define Y_BOTTOM_POS_PIN  A3
#define Y_BOTTOM_POS_ACTIVE LOW

#ifdef FIX_Z_DIST
  #define Z_STEP_PIN        3
  #define Z_STEP_ACTIVE     LOW
  #define Z_STEP_PASSIVE    HIGH
  #define Z_DIR_PIN         2
  #define Z_DIR_FARTHER     LOW
  #define Z_DIR_CLOSER      HIGH
  #define Z_CLOSER_POS_PIN  A4
  #define Z_CLOSER_POS_ACTIVE LOW
  #define Z_FARTHER_POS_PIN A5
  #define Z_FARTHER_POS_ACTIVE LOW
#endif

//int step_count_per_pixel = 8;
//int half_step_delay_mksek = 1000 / step_count_per_pixel;
//int stop_step_delay_mksek = half_step_delay_mksek * 10;

#pragma pack(1)
int global_step_delay_x = 0;
int global_step_delay_y = 0;
unsigned char global_send_state_b = 0;
unsigned long global_send_state_tm = 0;
long global_z = 0;
union {
  unsigned char raw[1];
  struct {
    unsigned long full_state;
    long x;
    long y;
    unsigned char cnt_per_pixel_x;
    unsigned char cnt_per_pixel_y;
    unsigned char stop_multi;
    unsigned char pwr_max;
    unsigned short step_delay;
  };
  struct {
    unsigned long inited_xy : 1;
    unsigned long inited_z : 2;
    unsigned long enabled : 1;
    unsigned long emergency : 1;
    unsigned long left : 1;
    unsigned long right : 1;
    unsigned long top : 1;
    unsigned long bottom : 1;
    unsigned long closer : 1;
    unsigned long farther : 1;    
    unsigned long is_moving : 1;
    unsigned long queue_dir : 1;
    unsigned long queue_len : 10;
    unsigned long inited_cfg : 1;
    unsigned long invert_x : 1;
    unsigned long invert_y : 1;
    unsigned long check_but : 1;
  };
  struct {
    unsigned long inited_all : 3;
    unsigned long en_state : 2;
    unsigned long lrtb_all : 4;
  };
} globals;

struct {  
  long x;
  long y;
} move_task;

unsigned char queue_pwr_data[1023];

int inc_proto_index = -1;
unsigned char inc_proto_esc = 0;
union {
  unsigned char raw[1];
  unsigned char cmd;
  struct {
    unsigned char cmd;    
    long x;
    long y;
  } move_cmd;
  struct {
    unsigned char cmd; 
    unsigned char dir;
    unsigned char data[256];   
  } queue_cmd;
  struct {
    unsigned char cmd;
    union {
      unsigned char cfg_byte;
      struct {
        unsigned char invert_x : 1;
        unsigned char invert_y : 1;
        unsigned char check_but : 1;
      };
    };
    unsigned char cnt_per_pixel_x;
    unsigned char cnt_per_pixel_y;
    unsigned char stop_multi;
    unsigned char pwr_max;
    unsigned short step_delay;
  } conf_cmd;
} inc_proto;
#pragma pack()

void proc_inc_pack() {
  switch (inc_proto.cmd) {
    case 1: // move
      if (inc_proto_index < 9) {
        break;
      }
      move_task.x = inc_proto.move_cmd.x;
      move_task.y = inc_proto.move_cmd.y;
      break;
    case 2: // queue
      if (inc_proto_index < 2) {
        break;
      }
      inc_proto_index -= 2;
      if (globals.queue_dir != inc_proto.queue_cmd.dir) {
        globals.queue_len = 0;
      }
      globals.queue_dir = inc_proto.queue_cmd.dir & 0x01;
      if (inc_proto_index + globals.queue_len > sizeof(queue_pwr_data)) {
        inc_proto_index = sizeof(queue_pwr_data) - globals.queue_len;
      }
      if (inc_proto_index > 0) {
        memmove(&queue_pwr_data[inc_proto_index], queue_pwr_data, globals.queue_len);
        memcpy(queue_pwr_data, inc_proto.queue_cmd.data, inc_proto_index);
        globals.queue_len += inc_proto_index;
      }
      break;
    case 3: // cfg
      if (inc_proto_index < 8) {
        break;
      }
      globals.cnt_per_pixel_x = inc_proto.conf_cmd.cnt_per_pixel_x;
      if (globals.cnt_per_pixel_x == 0) globals.cnt_per_pixel_x = 1;
      globals.cnt_per_pixel_y = inc_proto.conf_cmd.cnt_per_pixel_y;
      if (globals.cnt_per_pixel_y == 0) globals.cnt_per_pixel_y = 1;
      globals.stop_multi = inc_proto.conf_cmd.stop_multi;
      if (globals.stop_multi == 0) globals.stop_multi = 1;
      globals.pwr_max = inc_proto.conf_cmd.pwr_max;
      if (globals.pwr_max > 250) globals.pwr_max = 250;
      globals.step_delay = inc_proto.conf_cmd.step_delay;
      if (globals.step_delay < 100) globals.step_delay = 100;
      globals.invert_x = inc_proto.conf_cmd.invert_x;
      globals.invert_y = inc_proto.conf_cmd.invert_y;
      globals.check_but = inc_proto.conf_cmd.check_but;
      globals.inited_cfg = 1;
      globals.inited_xy = 0;
      global_send_state_b = 1;
      break;    
  }
}

void proc_inc_proto(unsigned char chr) {
  switch(chr) {
    case 0xAA:
      inc_proto_index = 0;
      inc_proto_esc = 0;
      break;
    case 0xAB:
      if (inc_proto_index >= 1) {
        proc_inc_pack();
      }
      inc_proto_index = -1;
      break;
    case 0xA0:
      inc_proto_esc = 1;
      break;
    default:
      if (inc_proto_index < 0) {
        break;
      }
      if (inc_proto_esc) {
        chr += 0xA0;
        inc_proto_esc = 0;
      }
      if (inc_proto_index >= sizeof(inc_proto)) {
        inc_proto_index = -1;
        break;
      }
      inc_proto.raw[inc_proto_index] = chr;        
      inc_proto_index++;
      break;
  }
}

inline long get_sign(long val) {
  if (val > 0) return 1;
  if (val < 0) return -1;
  return 0;
}
inline int get_abs(int val) {
  if (val < 0) return -val;
  return val;
}
inline long get_abs(long val) {
  if (val < 0) return -val;
  return val;
}

void do_x_step(long x_dir, int x_next) {
  unsigned char is_x = 0;
  if (x_dir > 0) {
    digitalWrite(X_DIR_PIN, globals.invert_x ? 1 : 0);
    if (globals.right == 0) { is_x = 1; }
  } else if (x_dir < 0) {
    digitalWrite(X_DIR_PIN, globals.invert_x ? 0 : 1);
    if (globals.left == 0) { is_x = 1; }
  }
  if (is_x == 0) {
    return;
  }
  int new_step_delay_x = get_sign(x_dir);
  if ((get_abs(x_dir) <= 1) && ((x_next == 0) || (get_sign(x_dir) != get_sign(x_next)))) {
    new_step_delay_x = 0;
  }
  int cur_delay = (int)globals.step_delay / (int)globals.cnt_per_pixel_x; //half_step_delay_mksek;
  int end_delay = cur_delay;
  if (global_step_delay_x == 0) {
    cur_delay *= (int)globals.stop_multi; //stop_step_delay_mksek;
  }
  if (new_step_delay_x == 0) {
    end_delay *= (int)globals.stop_multi; //stop_step_delay_mksek;
  }
  int delay_delta = (end_delay - cur_delay) / (int)globals.cnt_per_pixel_x; // step_count_per_pixel;
  for (int step_i = 0; step_i < (int)globals.cnt_per_pixel_x; step_i++) { // step_count_per_pixel
    delayMicroseconds(cur_delay);
    digitalWrite(X_STEP_PIN, LOW);
    delayMicroseconds(cur_delay);
    digitalWrite(X_STEP_PIN, HIGH);
    cur_delay += delay_delta;
  }
  globals.x += get_sign(x_dir);
  global_step_delay_x = new_step_delay_x;
}

void do_y_step(long y_dir) {
  unsigned char is_y = 0;
  if (y_dir > 0) {
    digitalWrite(Y_DIR_PIN, globals.invert_y ? 1 : 0);
    if (globals.bottom == 0) { is_y = 1; }
  } else if (y_dir < 0) {
    digitalWrite(Y_DIR_PIN, globals.invert_y ? 0 : 1);
    if (globals.top == 0) { is_y = 1; }
  }
  if (is_y == 0) {
    return;
  }
  int new_step_delay_y = get_sign(y_dir);
  if (get_abs(y_dir) <= 1) {
    new_step_delay_y = 0;
  }
  int cur_delay = (int)globals.step_delay / (int)globals.cnt_per_pixel_y; //half_step_delay_mksek;
  int end_delay = cur_delay;
  if (global_step_delay_y == 0) {
    cur_delay *= (int)globals.stop_multi; //stop_step_delay_mksek;
  }
  if (new_step_delay_y == 0) {
    end_delay *= (int)globals.stop_multi; //stop_step_delay_mksek;
  }
  int delay_delta = (end_delay - cur_delay) / (int)globals.cnt_per_pixel_y; // step_count_per_pixel;
  for (int step_i = 0; step_i < (int)globals.cnt_per_pixel_y; step_i++) { // step_count_per_pixel
    delayMicroseconds(cur_delay); // half_step_delay_mksek
    digitalWrite(Y_STEP_PIN, LOW);
    delayMicroseconds(cur_delay); // half_step_delay_mksek
    digitalWrite(Y_STEP_PIN, HIGH);
    cur_delay += delay_delta;
  }
  globals.y += get_sign(y_dir);
  global_step_delay_y = new_step_delay_y;
}

inline void do_z_step(long z_up) {
#ifdef FIX_Z_DIST
  digitalWrite(Z_DIR_PIN, z_up > 0 ? Z_DIR_FARTHER : Z_DIR_CLOSER);
  for (int step_i = 0; step_i < step_count_per_pixel; step_i++) {
    digitalWrite(Z_STEP_PIN, Z_STEP_ACTIVE);
    delayMicroseconds(half_step_delay_mksek);
    digitalWrite(Z_STEP_PIN, Z_STEP_PASSIVE);
    delayMicroseconds(half_step_delay_mksek);
  }
#endif
  global_z += get_sign(z_up);
}

/*
void do_power(int pwr) {
  if (pwr = 0) {
    digitalWrite(LASER_POWER_PIN, LASER_POWER_OFF);
  } else {
    if (pwr > 250) {
      pwr = 250;
    }
    analogWrite(LASER_POWER_PIN, pwr);
  }
}
inline void skip_serial() {
  while (Serial.available() > 0) { Serial.read(); }
}
*/

void write_ext_data(unsigned char * buf, int len) {
  int i;
  if (len <= 0) {
    return;
  }
  Serial.write(0xAA);
  for (i = 0; i < len; i++) {
      switch (buf[i]) {
      case 0xAA:
      case 0xAB:
      case 0xA0:
          Serial.write(0xA0);
          Serial.write(buf[i] - 0xA0);
          break;        
      default:
          Serial.write(buf[i]);
          break;
      }        
  }
  Serial.write(0xAB);
}

void send_state() {
  write_ext_data(globals.raw, sizeof(globals));
}

inline void check_send_state() {
  unsigned long cur_tm = millis();
  if ((unsigned long)(cur_tm - global_send_state_tm) > 100 || global_send_state_b) {
    global_send_state_tm = cur_tm;
    global_send_state_b = 0;
    send_state();
  }
}

int left_pin_count = 0;
int right_pin_count = 0;
int top_pin_count = 0;
int bottom_pin_count = 0;
#ifdef EMERGENCY_PIN
  int emergency_pin_count = 0;
#endif
inline void check_inputs() {
#ifdef EMERGENCY_PIN
  //globals.emergency = digitalRead(EMERGENCY_PIN) == EMERGENCY_ACTIVE ? 1 : 0;
  if (globals.check_but && digitalRead(EMERGENCY_PIN) == EMERGENCY_ACTIVE) {
    //globals.emergency = 1;
    if (++emergency_pin_count > PIN_FILTER_COUNT) {
      globals.emergency = 1;
    }
  } else {
    globals.emergency = 0;
    emergency_pin_count = 0;
  }
#else
  globals.emergency = 0;
#endif
  if (globals.check_but && digitalRead(X_LEFT_POS_PIN) == X_LEFT_POS_ACTIVE) {
    //globals.left = 1;
    if (++left_pin_count > PIN_FILTER_COUNT) {
      globals.left = 1;
    }
  } else {
    globals.left = 0;
    left_pin_count = 0;
  }
  if (globals.check_but && digitalRead(X_RIGHT_POS_PIN) == X_RIGHT_POS_ACTIVE) {
    //globals.right = 1;
    if (++right_pin_count > PIN_FILTER_COUNT) {
      globals.right = 1;
    }
  } else {
    globals.right = 0;
    right_pin_count = 0;
  }
  if (globals.check_but && digitalRead(Y_TOP_POS_PIN) == Y_TOP_POS_ACTIVE) { 
    //globals.top = 1;
    if (++top_pin_count > PIN_FILTER_COUNT) {
      globals.top = 1;
    }
  } else {
    globals.top = 0;
    top_pin_count = 0;
  }
  if (globals.check_but && digitalRead(Y_BOTTOM_POS_PIN) == Y_BOTTOM_POS_ACTIVE) {
    //globals.bottom = 1;
    if (++bottom_pin_count > PIN_FILTER_COUNT) {
      globals.bottom = 1;
    }
  } else {
    globals.bottom = 0;
    bottom_pin_count = 0;
  }
#ifdef FIX_Z_DIST
  globals.closer = digitalRead(Z_CLOSER_POS_PIN) == Z_CLOSER_POS_ACTIVE ? 1 : 0;
  globals.farther = digitalRead(Z_FARTHER_POS_PIN) == Z_FARTHER_POS_ACTIVE ? 1 : 0;
#else
  globals.closer = 0;
  globals.farther = 0;
#endif
}

inline unsigned char check_emergency() {
  if (globals.emergency) {
    digitalWrite(ALL_POWER_PIN, ALL_POWER_OFF);
    globals.enabled = 0;
    global_step_delay_x = 0;
    global_step_delay_y = 0;
    //skip_serial();
    return 1;
  } else if (globals.enabled == 0) {
    digitalWrite(ALL_POWER_PIN, ALL_POWER_ON);
    globals.enabled = 1;
  }
  return 0;
}

void setup() {
  memset(&globals, 0, sizeof(globals));
  //globals.inited_xy = 1;
  globals.inited_xy = 0; 
#ifdef FIX_Z_DIST
  globals.inited_z = 0;
#else
  globals.inited_z = 2;
#endif
  move_task.x = -1; 
  move_task.y = -1;
  globals.x = 10000;
  globals.y = 10000;
  global_step_delay_x = 0;
  global_step_delay_y = 0;

  globals.left = 0;
  left_pin_count = 0;
  globals.right = 0;
  right_pin_count = 0;
  globals.top = 0;
  top_pin_count = 0;
  globals.bottom = 0;
  bottom_pin_count = 0;

  globals.cnt_per_pixel_x = 8;
  globals.cnt_per_pixel_y = 8;
  globals.stop_multi = 10;
  globals.pwr_max = 200;
  globals.step_delay = 1000;
  globals.invert_x = 0;
  globals.invert_y = 0;
  globals.check_but = 0;
  globals.inited_cfg = 0;
  
  Serial.begin(115200);
#ifdef FIX_Z_DIST
  pinMode(Z_DIR_PIN, OUTPUT); digitalWrite(Z_DIR_PIN, Z_DIR_CLOSER);
  pinMode(Z_STEP_PIN, OUTPUT); digitalWrite(Z_STEP_PIN, Z_STEP_PASSIVE);
  pinMode(Z_CLOSER_POS_PIN, INPUT_PULLUP);
  pinMode(Z_FARTHER_POS_PIN, INPUT_PULLUP);
#endif
  pinMode(X_DIR_PIN, OUTPUT); digitalWrite(X_DIR_PIN, HIGH);
  pinMode(X_STEP_PIN, OUTPUT); digitalWrite(X_STEP_PIN, HIGH);
  pinMode(Y_DIR_PIN, OUTPUT); digitalWrite(Y_DIR_PIN, HIGH);
  pinMode(Y_STEP_PIN, OUTPUT); digitalWrite(Y_STEP_PIN, HIGH);
  pinMode(ALL_POWER_PIN, OUTPUT); digitalWrite(ALL_POWER_PIN, ALL_POWER_OFF);
  pinMode(LASER_POWER_PIN, OUTPUT); digitalWrite(LASER_POWER_PIN, LASER_POWER_OFF); setPwmFrequency(LASER_POWER_PIN, 8); // 8 => 4 kHz
  pinMode(X_LEFT_POS_PIN, INPUT_PULLUP);
  pinMode(X_RIGHT_POS_PIN, INPUT_PULLUP);
  pinMode(Y_TOP_POS_PIN, INPUT_PULLUP);
  pinMode(Y_BOTTOM_POS_PIN, INPUT_PULLUP);
#ifdef EMERGENCY_PIN
  pinMode(EMERGENCY_PIN, INPUT_PULLUP);
#endif
  global_send_state_tm = millis();
  global_send_state_b = 1;
}

void loop_initial() {
  //skip_serial();
  if (globals.check_but && globals.top == 0) {
    do_y_step(-1);
  } else if (globals.check_but && globals.left == 0) {
    do_x_step(-1, -100);
  } else {
    globals.inited_xy = 1;
    global_send_state_b = 1;
    globals.x = 0;
    globals.y = 0;
  }
#ifdef FIX_Z_DIST
  if ((globals.closer == 0) && (globals.inited_z == 0)) {
    do_z_step(-1);
  } else if (globals.inited_z == 0) {
    globals.inited_z = 1;
    global_send_state_b = 1;
    global_z = 0;
  } else if (FIX_Z_DIST != global_z) {    
    do_z_step(FIX_Z_DIST - global_z);
  } else {
    globals.inited_z = 2;
    global_send_state_b = 1;
  }
#else
  if (globals.inited_z != 2) {
    globals.inited_z = 2;
  }
#endif
}

inline void check_serial() {
  while (Serial.available() > 0) { proc_inc_proto(Serial.read()); }
}

void loop_work() {
  unsigned char make_stop_step = 1;
  //check_serial();
  if (globals.is_moving) {
    if (move_task.x >= 0 && move_task.y >= 0 && (move_task.x != globals.x || move_task.y != globals.y)) {
      //do_xy_step(move_task.x - globals.x, move_task.y - globals.y, globals.queue_dir ? (int)globals.queue_len : -(int)globals.queue_len);      
      if (move_task.y != globals.y) {
        do_y_step(move_task.y - globals.y);      
      } else if (move_task.x != globals.x) {
        do_x_step(move_task.x - globals.x, globals.queue_dir ? (int)globals.queue_len : -(int)globals.queue_len);      
      }
      return;
    } else {
      globals.is_moving = 0;
      move_task.x = -1;
      move_task.y = -1;
      global_send_state_b = 1;
    }
  } 
  
  if (globals.queue_len > 0) {
    globals.queue_len--;
    if (globals.lrtb_all == 0) { 
      //do_power(queue_pwr_data[globals.queue_len]);
      int pwr = queue_pwr_data[globals.queue_len];
      if (pwr > (int)globals.pwr_max) {
        analogWrite(LASER_POWER_PIN, globals.pwr_max);
      } else {
        analogWrite(LASER_POWER_PIN, pwr);
      }
    } else {
      //do_power(0);
      digitalWrite(LASER_POWER_PIN, 0);
    }    
    //do_xy_step(globals.queue_dir ? 1 : -1, 0, globals.queue_dir ? 1024 : -1024); make_stop_step = 0;
    do_x_step(globals.queue_dir ? 1 : -1, globals.queue_dir ? 1024 : -1024); make_stop_step = 0;
    if (globals.queue_len == 0) { 
      //do_power(0); 
      digitalWrite(LASER_POWER_PIN, 0);
      global_send_state_b = 1;
    } else {
      return;
    }
  } 
  if (move_task.x >= 0 && move_task.y >= 0) {
    globals.is_moving = 1;
    global_send_state_b = 1;
    //do_xy_step(move_task.x - globals.x, move_task.y - globals.y, globals.queue_dir ? (int)globals.queue_len : -(int)globals.queue_len);      
    if (move_task.y != globals.y) {
      do_y_step(move_task.y - globals.y);      
    } else if (move_task.x != globals.x) {
      do_x_step(move_task.x - globals.x, globals.queue_dir ? (int)globals.queue_len : -(int)globals.queue_len);      
    }
  } else if (make_stop_step) {
    //do_xy_step(0, 0, 0);      
  }
}

void loop() {
  check_inputs();
  check_send_state();
  check_serial();
  if (check_emergency()) { return; }
  if (globals.inited_cfg == 0) {    
  } else if (globals.inited_xy != 1 || globals.inited_z != 2) {
    loop_initial();
  } else {
    loop_work();
  }
}

/*
 * Посмотрел документацию на драйвера – я правильно понимаю, что энкодер от двигателя будет подключаться напрямую к драйверу, и он сам будет рулить пропуском шагов?
 Да все правильно, драйвер сам рулит этой системой.
И по той же документации на драйвера, как я понимаю, для входов STEP, DIR и ENABLE используется оптоэлектронная развязка, соответственно один вход идет на +3/+5 вольт, а второй – на управление контроллером, итого получается 3 провода на канал, причем ENABLE от всех осей можно объединить в один, и объединить с включение обдува лазера (через твердотельное реле). 
Это я не к тому, что надо использовать разъем с меньшим количеством линий, а к выбору правильной ардуины, то есть в том разъеме, который идет к ардуине можно будет часть линий объединить, а часть пустить на +5 вольт.
 Тут главное правило все питания и земли разводятся по типу звезды от источника питания. + провод рекомендуется использовать на канал витая пара. Enable идет переплетен со своим питание Dir со своим и тд. Все это в экране. Максимально хочется ограничиться от влияния помех и глюков при работе.  На выбор ардуины это никак не повлияет, просто соединим питания до разъема как будет нужно. (но от разъема пойдет 6 проводов по 3 пары) на каждый драйвер.
По поводу Enable думаю что нужно предусмотреть режим  когда необходимо выключить какойто шаговик и не лезть откручивать питание от него. Я бы сделал их раздельными на оси Х т.к. вдруг придется выравнивать портал то выключив один драйвер вторым мы исправляем перекос. (ось X управляется одинаковыми сигналами Step и Dir поэтому они будут шагать одинаково и для выравнивания нужно один отключать)
Во всех остальных их можно объединить в один.
Итого, что у меня получается по ардуине:
1)      +5 питание
2)      Земля
3)      STEP X (выход)
4)      STEP Y (выход)
5)      STEP Z (выход)
6)      DIR X (выход)
7)      DIR Y (выход)
8)      DIR Z (выход)
9)      ENABLE X + ENABLE Y + ENABLE Z + включение обдува + включения всего остального что необходимо (выход)
10)  LASER PWM (выход)
11)  X LEFT + X RIGHT (концевики по краевому положению на оси X – можно объединить если они срабатывают на замыкание) (вход)
12)  Y LEFT + Y RIGHT (вход)
13)  Z LEFT + Z RIGHT (вход)
14)  EMERGENCY STOP + защита от вскрытия + все остальное по чему надо резко остановиться (можно объединить если они срабатывают все на замыкание) (вход)
ПоП
 По концевикам их объединять нельзя. Первый на оси это начальное положение (Home) второй (конец оси) если вдруг выйдет по каким то причинам за пределы рабочего поля должен произойти программный стоп с сообщением об ошибке.
Если по какойто причине програмно контроллер не остановил ось должен сработать аварийный концевик (он стоит дальше ) и выключить питание двигателей физически их обесточить. Это тоже не повлияет на выбор ардуины. Это будем делать по месту.

 */
